# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import copy
import logging
import os
import sys
from unittest import main
from unittest.mock import MagicMock

from mlcvzoo_base.utils.file_utils import get_file_list
from pytest import fixture, mark
from pytest_mock import MockerFixture
from test_template import TestTemplate

from mlcvzoo_util.video_image_creator.video_image_creator import VideoImageCreator
from mlcvzoo_util.video_image_creator.video_image_creator import (
    main as video_image_creator_main,
)

logger = logging.getLogger(__name__)


@fixture(scope="function")
def cv2_waitKey_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "cv2.waitKey",
        return_value=ord("s"),
    )


@fixture(scope="function")
def cv2_named_window(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "cv2.namedWindow",
        return_value=None,
    )


@fixture(scope="function")
def cv2_resize_window(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "cv2.resizeWindow",
        return_value=None,
    )


@fixture(scope="function")
def cv2_imshow(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "cv2.imshow",
        return_value=None,
    )


class TestVideoImageCreator(TestTemplate):
    @mark.usefixtures(
        "cv2_waitKey_mock", "cv2_named_window", "cv2_resize_window", "cv2_imshow"
    )
    def test_video_image_creator_video_path(self) -> None:
        video_image_creator = VideoImageCreator(
            configuration=VideoImageCreator.create_configuration(
                yaml_config_path=os.path.join(
                    self.project_root,
                    "test_data",
                    "test_video_image_creator",
                    "test_video-image-creator_video-path.yaml",
                ),
                string_replacement_map=self.string_replacement_map,
            )
        )
        video_image_creator.run()

        written_video_images = get_file_list(
            input_dir=os.path.join(self.project_root, "test_data/test_video"),
            file_extension=".jpg",
        )

        assert len(written_video_images) == 10

    @mark.usefixtures(
        "cv2_waitKey_mock", "cv2_named_window", "cv2_resize_window", "cv2_imshow"
    )
    def test_video_image_creator_video_dir(self) -> None:
        video_image_creator = VideoImageCreator(
            configuration=VideoImageCreator.create_configuration(
                yaml_config_path=os.path.join(
                    self.project_root,
                    "test_data",
                    "test_video_image_creator",
                    "test_video-image-creator_video-dir.yaml",
                ),
                string_replacement_map=self.string_replacement_map,
            )
        )
        video_image_creator.run()

        written_video_images = get_file_list(
            input_dir=os.path.join(self.project_root, "test_data/test_video"),
            file_extension=".jpg",
        )

        assert len(written_video_images) == 10

    @mark.usefixtures(
        "cv2_waitKey_mock", "cv2_named_window", "cv2_resize_window", "cv2_imshow"
    )
    def test_video_image_creator_no_video_files(self) -> None:
        with self.assertRaises(ValueError):
            VideoImageCreator(
                configuration=VideoImageCreator.create_configuration(
                    yaml_config_path=os.path.join(
                        self.project_root,
                        "test_data",
                        "test_video_image_creator",
                        "test_video-image-creator_no-video-files.yaml",
                    ),
                    string_replacement_map=self.string_replacement_map,
                )
            )

    @mark.usefixtures(
        "cv2_waitKey_mock", "cv2_named_window", "cv2_resize_window", "cv2_imshow"
    )
    def test_video_image_creator_main(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_video_image_creator/",
                    "test_video-image-creator_video-path.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
            ]
        )

        video_image_creator_main()

        sys.argv = argv_copy


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for handling the upload of zip files to CVAT via their commandline interface"""

import logging
import os
import zipfile
from typing import Dict

import yaml
from mlcvzoo_base.utils import ensure_dir, get_file_list
from related import to_yaml

from mlcvzoo_util.cvat_annotation_handler.configuration import (
    CVATCLIConfig,
    CVATTaskInfoConfig,
    CVATTaskUploadConfig,
)
from mlcvzoo_util.cvat_annotation_handler.utils import run_command_in_process

logger = logging.getLogger(__name__)


class PascalVOCUploader:
    """
    Class for handling the upload of zip files to CVAT via their commandline-interface
    """

    @staticmethod
    def upload_task_data(
        upload_task_config: CVATTaskUploadConfig,
        cvat_cli_config: CVATCLIConfig,
        disable_ssl_verify: bool = False,
    ) -> None:
        """
        Execute the upload of a single zip file to the CVAT instance

        Args:
            upload_task_config: Configuration of the task that is to be uploaded
            cvat_cli_config: Basis configuration that is needed for the communication between
                             the commandline-interface and CVAT
            disable_ssl_verify: Whether to allow uncertified https connections

        Returns:
            None
        """

        logger.info(
            "\n========== UPLOAD TASK ==========\n" "Task config: \n" "'''\n%s'''\n",
            to_yaml(
                upload_task_config,
                yaml_package=yaml,
                dumper_cls=yaml.Dumper,
            ),
        )

        if upload_task_config.use_prediction_data:
            target_zip_path = PascalVOCUploader.write_upload_zip_file(
                upload_task_config=upload_task_config
            )
        else:
            target_zip_path = upload_task_config.source_zip_path
            upload_task_config.target_zip_path = target_zip_path

        if upload_task_config.execute_upload:
            PascalVOCUploader.call_cvat_upload_via_cli(
                cvat_cli_config=cvat_cli_config,
                target_zip_path=target_zip_path,
                task_info=upload_task_config.task_info,
                disable_ssl_verify=disable_ssl_verify,
            )

        if upload_task_config.clean_up_upload_data:
            PascalVOCUploader.__post_clean_up(
                source_zip_path=upload_task_config.source_zip_path,
                target_zip_path=target_zip_path,
            )

    @staticmethod
    def __post_clean_up(source_zip_path: str, target_zip_path: str) -> None:
        if os.path.isfile(source_zip_path):
            logger.info("(UPLOAD) Clean up dump zip file: '%s'", source_zip_path)
            os.remove(source_zip_path)

        if os.path.isfile(target_zip_path):
            logger.info("(UPLOAD) Clean up upload zip file: '%s'", target_zip_path)
            os.remove(target_zip_path)

    @staticmethod
    def call_cvat_upload_via_cli(
        cvat_cli_config: CVATCLIConfig,
        target_zip_path: str,
        task_info: CVATTaskInfoConfig,
        disable_ssl_verify: bool = False,
    ) -> None:
        """
        Upload a zip file to CVAT.

        Args:
            cvat_cli_config: Basis configuration that is needed for the communication between
                             the commandline-interface and CVAT
            target_zip_path: The zip file to upload
            task_info: Configuration for the specific task
            disable_ssl_verify: Whether to allow uncertified https connections

        Returns:
            None
        """

        # ================================
        # UPLOAD TASK DATA
        command = PascalVOCUploader.__create_cvat_upload_cli_command(
            base_command=cvat_cli_config.create_base_command_string(),
            target_zip_path=target_zip_path,
            task_info=task_info,
        )
        logger.info(
            "Upload annotation data: Task-ID=%s, " "zip-path=%s",
            task_info.task_ID,
            target_zip_path,
        )
        return_code = run_command_in_process(
            command=command, disable_ssl_verify=disable_ssl_verify
        )

        if return_code != 0:
            raise ValueError(f"Upload did not succeed, return-code={return_code}")

    @staticmethod
    def __create_cvat_upload_cli_command(
        base_command: str, target_zip_path: str, task_info: CVATTaskInfoConfig
    ) -> str:
        return (
            f"{base_command} upload "
            f"{task_info.task_ID} "
            f"{target_zip_path} "
            f"--format '{task_info.annotation_format}'"
        )

    @staticmethod
    def write_upload_zip_file(upload_task_config: CVATTaskUploadConfig) -> str:
        """
        Create a zip file based on generated PASCAL-VOC xml files,
        that can be uploaded to a CVAT instance.

        Args:
            upload_task_config: Configuration that states where to find the prediction data
                                and where the zip file should be saved

        Returns:
            The path to the zip file that has been written to disk
        """

        # Generate path to the zip-file which should be uploaded to CVAT
        if upload_task_config.target_zip_path == "":
            target_zip_path = upload_task_config.source_zip_path.replace(
                ".zip", "_upload.zip"
            )
        else:
            target_zip_path = upload_task_config.target_zip_path

        # CLEAN UP old upload zip-file
        if os.path.isfile(target_zip_path):
            logger.info("(UPLOAD) Clean up zip file: '%s'", target_zip_path)
            os.remove(target_zip_path)

        return PascalVOCUploader.__write_prediction_data_to_zip(
            source_zip_path=upload_task_config.source_zip_path,
            target_zip_path=target_zip_path,
            prediction_data_dir=upload_task_config.prediction_data_dir,
        )

    @staticmethod
    def __write_prediction_data_to_zip(
        source_zip_path: str, target_zip_path: str, prediction_data_dir: str
    ) -> str:
        if not os.path.isdir(prediction_data_dir):
            raise ValueError(
                f"prediction_data_dir='{prediction_data_dir}' does not exist! "
                "Please specify a directory in order to generate"
                "a prediction zip file for the upload to CVAT!"
            )

        # Open zip-file to write data which will be uploaded to the CVAT task
        ensure_dir(file_path=target_zip_path, verbose=True)
        target_zip_file = zipfile.ZipFile(  # pylint: disable=consider-using-with
            target_zip_path, "w"
        )

        # Open original zip file to read content that shouldbe copied
        source_zip_file = zipfile.ZipFile(
            source_zip_path
        )  # pylint: disable=consider-using-with
        original_zip_paths = zipfile.ZipFile.namelist(source_zip_file)
        for file_path in original_zip_paths:
            # Copy the 'labelmap.txt' of the original zip file to the one
            # that is uploaded
            if file_path == "labelmap.txt":
                target_zip_file.writestr(
                    zinfo_or_arcname="labelmap.txt",
                    data=source_zip_file.read("labelmap.txt"),
                )

            # Copy every file of the 'ImageSets' folder of the original zip file,
            # to the zip file that is uploaded
            if file_path.split("/")[0] == "ImageSets":
                target_zip_file.writestr(
                    zinfo_or_arcname=file_path, data=source_zip_file.read(file_path)
                )

        # NOTE:
        # - The annotations are defined in the download cvat zip file
        # - Therefore it is enough to specify a root directory where
        #   all xml files can be parsed from.
        # - The decision which xml files should be put in the upload
        #   zip file, is made on the basis of the information from the
        #   downloaded zip file from "ImageSets/Main/default.txt"
        predicted_xml_paths = get_file_list(
            input_dir=prediction_data_dir,
            file_extension=".xml",
            search_subfolders=True,
        )

        predicted_xml_paths_map: Dict[str, str] = {}
        for predicted_xml_path in predicted_xml_paths:
            predicted_xml_paths_map[
                os.path.basename(predicted_xml_path).replace(".xml", "")
            ] = predicted_xml_path

        # Get all xml paths that are part of the task (represented by the zip)
        annotation_xml_files = (
            source_zip_file.read("ImageSets/Main/default.txt")
            .decode("utf-8")
            .split("\n")
        )

        # Write annotation data to zip-file
        for annotation_xml_path in annotation_xml_files:
            annotation_file_name = os.path.basename(annotation_xml_path)

            if annotation_file_name in predicted_xml_paths_map:
                logger.debug(
                    "Add xml file to upload zip-file: '%s'",
                    predicted_xml_paths_map[annotation_file_name],
                )
                with open(
                    file=predicted_xml_paths_map[annotation_file_name], encoding="utf8"
                ) as prediction_file:
                    target_zip_file.writestr(
                        zinfo_or_arcname=f"Annotations/{annotation_xml_path}.xml",
                        data="".join(prediction_file.readlines()),
                    )

        source_zip_file.close()
        target_zip_file.close()

        logger.info(
            "Generated prediction zip file '%s' for upload to cvat.", target_zip_path
        )

        return target_zip_path



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for extracting frames from videos. In the context of the mlcvzoo, this
is used to build training datasets.
"""

import argparse
import copy
import json
import logging
import os
from datetime import datetime
from typing import Dict, List, Optional, cast

import cv2
import numpy as np
from config_builder import ConfigBuilder
from mlcvzoo_base.utils.file_utils import ensure_dir, get_file_list

from mlcvzoo_util.logger import Logger
from mlcvzoo_util.mlcvzoo_cli_tool import MLCVZooCLITool
from mlcvzoo_util.video_image_creator.configuration import VideoImageCreatorConfig

logger = logging.getLogger(__name__)


class VideoImageCreator(MLCVZooCLITool[VideoImageCreatorConfig]):
    """
    Tool to step through video and manually sort out frames to use as training data.
    When executed, displays every frame in given video separately. User is able to
    execute following commands:
    - Exit: 'q', 'Q', 'Esc'
    - Save current frame as .jpg: 's'
    - Change step size: '1' for 1 step, '2' for step, ...

    (See ./config/templates/tools/video-image-creator_template.yaml for example)
    """

    MODES: List[str] = ["terminate", "run"]

    def __init__(
        self,
        configuration: VideoImageCreatorConfig,
    ) -> None:
        self.video_files: List[str] = []

        self.current_pos: int = 0
        self.current_frame: Optional[np.ndarray] = None  # type: ignore[type-arg]
        self.current_video_capture: Optional[cv2.VideoCapture] = None
        self.current_video_path: Optional[str] = None

        self.configuration: VideoImageCreatorConfig = configuration

        self.step_width_map: Dict[int, int] = {}
        for key in self.configuration.step_width_map.keys():
            self.step_width_map[ord(key)] = self.configuration.step_width_map[key]

        self.mode = "run"
        self.frame = 0
        self.resized_window = False
        self.step_width = 1

        cv2.namedWindow(self.configuration.winname, cv2.WINDOW_NORMAL)

        self.video_files = []
        if os.path.isdir(self.configuration.video_input_dir):
            self.video_files.extend(
                get_file_list(
                    input_dir=self.configuration.video_input_dir,
                    search_subfolders=True,
                    file_extension=self.configuration.video_file_extension,
                )
            )
        elif os.path.isfile(self.configuration.video_input_path):
            self.video_files.append(self.configuration.video_input_path)
        else:
            logger.error(
                "Could not init a video file. Please provide a correct parameter for "
                " 'video_input_dir' or 'video_input_path'"
            )

        if len(self.video_files) == 0:
            raise ValueError("No videos found for the given configuration!")
        else:
            logger.info("Input Video: ")
            self.__print_videos()

    @staticmethod
    def create_configuration(
        yaml_config_path: str,
        string_replacement_map: Optional[Dict[str, str]] = None,
        no_checks: bool = False,
    ) -> VideoImageCreatorConfig:
        """
        Create a VideoImageCreatorConfig

        Args:
            yaml_config_path:  A yaml filepath where to build the configuration
                               object from
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
            no_checks: Whether the configuration object should be checked for mutual exclusiveness
                       and the "check_values" method for each attribute of the supertype
                       "BaseConfigClass" should be called
        Returns:
            The created configuration
        """

        return cast(
            VideoImageCreatorConfig,
            ConfigBuilder(
                class_type=VideoImageCreatorConfig,
                yaml_config_path=yaml_config_path,
                string_replacement_map=string_replacement_map,
                no_checks=no_checks,
            ).configuration,
        )

    @staticmethod
    def create_cli_configuration(
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> VideoImageCreatorConfig:
        """
        Create a VideoImageCreatorConfig

        Args:
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
        Returns:
            The created configuration
        """

        return MLCVZooCLITool._create_cli_configuration(
            configuration_class=VideoImageCreatorConfig,
            string_replacement_map=string_replacement_map,
            configure_argparse=VideoImageCreator.configure_argparse,
        )

    @staticmethod
    def configure_argparse(
        parser: argparse.ArgumentParser,
    ) -> None:
        MLCVZooCLITool.configure_argparse(parser)

        parser.description = (
            "Tool for generating training images from given video files"
        )

    def __print_videos(self) -> None:
        """
        Prints every video path specified in video_image_creator_yaml
        """
        for video_path in self.video_files:
            logger.info("  - %s", video_path)

    def __output_frame(self, frame: Optional[np.ndarray]) -> bool:  # type: ignore[type-arg]
        """
        Displays current frame incl. information about video name,
        current frame number and video FPS. While displaying frame,
        user can change step size according to step_width_map provided
        in video_image_creator_config_yaml. Also, user can save the current
        frame as single .jpg by pressing 's'

        Args:
            frame: N-dimensional Array representing single frame of video

        Returns:
            modified_value: Boolean which indicates if step size has changed

        """

        modified_value = False

        if (
            self.current_video_path is not None
            and self.current_video_capture is not None
            and frame is not None
        ):
            output_frame = copy.deepcopy(frame)

            # will only be executed once
            if not self.resized_window:
                scale_factor = frame.shape[1] / frame.shape[0]

                if self.configuration.resize_window is True:
                    window_height = self.configuration.window_size
                else:
                    window_height = frame.shape[0]

                cv2.resizeWindow(
                    winname=self.configuration.winname,
                    height=window_height,
                    width=int(window_height * scale_factor),
                )

                self.resized_window = True

            output_frame = cv2.putText(
                img=output_frame,
                text=f"Video: {os.path.basename(self.current_video_path)}",
                org=(20, 40),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=1.25,
                color=(0, 0, 0),
                thickness=3,
                lineType=cv2.LINE_AA,
            )

            output_frame = cv2.putText(
                img=output_frame,
                text=f"Current frame number: {self.current_pos}/"
                f"{self.current_video_capture.get(cv2.CAP_PROP_FRAME_COUNT)}",
                org=(20, 80),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=1.25,
                color=(0, 0, 0),
                thickness=3,
                lineType=cv2.LINE_AA,
            )

            output_frame = cv2.putText(
                img=output_frame,
                text=f"Video FPS: {self.current_video_capture.get(cv2.CAP_PROP_FPS)}",
                org=(20, 120),
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=1.25,
                color=(0, 0, 0),
                thickness=3,
                lineType=cv2.LINE_AA,
            )

            cv2.imshow(self.configuration.winname, output_frame)

            key = cv2.waitKey(0) & 0xFF
            if key in self.step_width_map.keys():
                logger.info("set step-width to %s", self.step_width_map[key])
                self.step_width = self.step_width_map[key]

                modified_value = True

            elif key == 27 or key == ord("q") or key == ord("Q"):
                self.close()
            elif key == ord("s"):
                logger.info("save image ...")
                self.__write_frame(frame=frame)
            else:
                logger.info("next image...")

            self.frame += 1

            del output_frame

        return modified_value

    def __write_frame(self, frame: Optional[np.ndarray]) -> None:  # type: ignore[type-arg]
        """
        Saves given frame as .jpg to same directory as input video

        Args:
            frame: N-dimensional Array representing single frame of video

        Returns:
            None
        """

        if (
            frame is not None
            and self.current_video_path is not None
            and self.current_video_capture is not None
        ):
            video_fps = self.current_video_capture.get(cv2.CAP_PROP_FPS)
            video_pos = self.current_video_capture.get(cv2.CAP_PROP_POS_FRAMES)

            timestamp = datetime.utcfromtimestamp(1 / video_fps * video_pos).strftime(
                "%H-%M-%S-%f"
            )

            video_name_base = os.path.basename(self.current_video_path).replace(
                self.configuration.video_file_extension, ""
            )

            output_path = os.path.join(
                os.path.dirname(self.current_video_path),
                video_name_base,
                f"{video_name_base}_{timestamp}.jpg",
            )

            if not os.path.isfile(output_path):
                ensure_dir(file_path=output_path, verbose=True)

                logger.info("Write image to %s", output_path)
                cv2.imwrite(output_path, frame, [cv2.IMWRITE_JPEG_QUALITY, 100])
            else:
                logger.info("Image already exist at %s", output_path)

    def __reset(self) -> None:
        """
        Closes current video and resets current_video variables to None
        """

        if self.current_video_capture is not None:
            self.current_video_capture.release()

        self.current_video_capture = None
        self.current_video_path = None

        logger.info("Reset video. Videos left: ")
        self.__print_videos()

    def run(self) -> None:
        """
        Wrapper function for executing the video_image_creator
        """

        modified_value = False

        logger.info(
            "Press 'Esc', 'q' or 'Q' to exit.\n"
            "Press 's' for saving current frame.\n"
            "step_width_map: \n"
            "%s\n"
            "Press any other key to step through video.",
            json.dumps(obj=self.step_width_map, indent=2),
        )
        while self.mode != "terminate":
            if self.current_video_capture is None:
                if len(self.video_files) == 0:
                    self.mode = "terminate"
                    break
                else:
                    self.current_video_path = self.video_files.pop()

                    logger.info("Start with video %s", self.current_video_path)
                    self.current_video_capture = cv2.VideoCapture(
                        self.current_video_path
                    )

            if not modified_value:
                self.current_pos = (
                    int(self.current_video_capture.get(cv2.CAP_PROP_POS_FRAMES))
                    + self.step_width
                    - 1
                )
                self.goto_frame(frame_position=self.current_pos)

                logger.info(
                    "Read frame at position: %s",
                    self.current_video_capture.get(cv2.CAP_PROP_POS_FRAMES),
                )
                read, self.current_frame = self.current_video_capture.read()

                if not read:
                    logger.info("Video finished. Read new video ...")
                    self.__reset()
                    continue
            else:
                logger.info("stay at pos")

            modified_value = self.__output_frame(frame=self.current_frame)

    def goto_frame(self, frame_position: int) -> None:
        """
        Takes integer specifying which frame should be displayed
        next and sets the frame accordingly.

        Args:
            frame_position: Integer specifying which frame_position to be set next

        Returns:
            None
        """

        if self.current_video_capture is not None:
            frame_position = max(0, frame_position)

            logger.info("set frame position to %s", frame_position)
            self.current_video_capture.set(cv2.CAP_PROP_POS_FRAMES, frame_position)

    def __del__(self):  # type: ignore
        if self.current_video_capture is not None:
            self.current_video_capture.release()

    def close(self) -> None:
        """
        Destroys the current window
        """

        self.mode = "terminate"
        cv2.destroyAllWindows()


def main() -> None:
    """
    Entry point when using the video-image-creator command line tool.
    (See [tool.poetry.scripts] section in pyproject.toml)
    """

    args = ConfigBuilder.setup_argparse(
        configure_argparse=VideoImageCreator.configure_argparse
    ).parse_args()

    Logger.init_logging_basic(
        log_dir=args.log_dir,
        log_file_postfix="VideoImageCreator",
        no_stdout=False,
        root_log_level=args.log_level,
    )

    video_image_creator = VideoImageCreator(
        configuration=VideoImageCreator.create_cli_configuration()
    )
    video_image_creator.run()


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module for storing utility functions that are needed for the cvat_annotation_handler package"""

import logging
import os
import shlex
import subprocess
import sys

logger = logging.getLogger(__name__)


def run_command_in_process(command: str, disable_ssl_verify: bool = False) -> int:
    """
    Run the CVAT commandline-interface in subprocess

    Args:
        command: The concrete command to be executed
        disable_ssl_verify: Whether to allow uncertified https connections

    Returns:

    """
    env = os.environ.copy()
    env["PYTHONPATH"] = os.pathsep.join(sys.path)
    env["PYTHONPATH"] += os.pathsep + os.getcwd()

    if disable_ssl_verify:
        env["CURL_CA_BUNDLE"] = ""

    logger.info(f"Run command: {command}")

    result = subprocess.run(shlex.split(command), env=env)
    logger.info(f"Executed command: {command}, result.returncode: {result.returncode}")

    return result.returncode



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""Module that provides a generic component for measuring runtimes of mlcvzoo models"""

import argparse
import logging
import time
from typing import Any, Dict, List, Optional, Type, cast

import cv2
import mlflow
from config_builder import ConfigBuilder
from mlcvzoo_base.api.model import Model
from mlcvzoo_base.configuration.model_config import ModelConfig
from mlcvzoo_base.configuration.replacement_config import ReplacementConfig
from mlcvzoo_base.configuration.structs import MLFlowExperimentTypes
from mlcvzoo_base.metrics.mlflow.mlflow_runner import MLFLowRunner
from mlcvzoo_base.models.model_registry import ModelRegistry
from mlcvzoo_base.utils.gpu_util import GpuInfo, get_device_info

from mlcvzoo_util.logger import Logger
from mlcvzoo_util.mlcvzoo_cli_tool import MLCVZooCLITool, configure_model_argparse
from mlcvzoo_util.model_timer.configuration import ModelTimerConfig

logger = logging.getLogger(__name__)


class ModelTimer(MLCVZooCLITool[ModelTimerConfig]):
    """
    Utility class for measuring the individual inference time of a group of models
    specified in the ModelTimer config file and logging the results to MLFlow.
    """

    def __init__(self, configuration: ModelTimerConfig) -> None:
        self.configuration: ModelTimerConfig = configuration

        self.mlflow_runner: Optional[MLFLowRunner] = None
        if self.configuration.mlflow_config is not None:
            self.mlflow_runner = MLFLowRunner(
                configuration=self.configuration.mlflow_config,
            )
        else:
            logger.info(
                "Not able to create MLFLowRunner since no MLFlowConfig was "
                "set in ModelTimerConfig.mlflow_config. "
                "Just logging execution time information to standard output."
            )

        self.inference_time_mean: Optional[float] = None
        self.inference_time_list: List[float] = []

    @staticmethod
    def create_configuration(
        yaml_config_path: str,
        string_replacement_map: Optional[Dict[str, str]] = None,
        no_checks: bool = False,
    ) -> ModelTimerConfig:
        """
        Create a ModelTimerConfig

        Args:
            yaml_config_path:  A yaml filepath where to build the configuration
                               object from
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
            no_checks: Whether the configuration object should be checked for mutual exclusiveness
                       and the "check_values" method for each attribute of the supertype
                       "BaseConfigClass" should be called
        Returns:
            The created configuration
        """

        return cast(
            ModelTimerConfig,
            ConfigBuilder(
                class_type=ModelTimerConfig,
                yaml_config_path=yaml_config_path,
                string_replacement_map=string_replacement_map,
                no_checks=no_checks,
            ).configuration,
        )

    @staticmethod
    def create_cli_configuration(
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> ModelTimerConfig:
        """
        Create a ModelTimerConfig

        Args:
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
        Returns:
            The created configuration
        """

        return MLCVZooCLITool._create_cli_configuration(
            configuration_class=ModelTimerConfig,
            string_replacement_map=string_replacement_map,
            configure_argparse=ModelTimer.configure_argparse,
        )

    @staticmethod
    def configure_argparse(
        parser: argparse.ArgumentParser,
    ) -> None:
        parser.description = (
            "Tool for measuring the individual inference time of a "
            "group of models specified in the ModelTimer config file "
            "and optionally log the results to MLFlow."
        )
        MLCVZooCLITool.configure_argparse(parser)
        configure_model_argparse(parser=parser)

    @staticmethod
    def create_model(
        model_config: ModelConfig,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> Model:  # type: ignore[type-arg]
        """
        Creates a model based on the given config

        Args:
            model_config: The model config that should be used
                          for creating a model
            string_replacement_map: (Optional) A dictionary that defines placeholders which can
                                    be used while parsing a configuration file. They can be
                                    understood as variables that can be used to define configs
                                    that are valid across multiple devices.
        Returns:
            The created model
        """
        # In order to be able to load the checkpoints that
        # should be evaluated, we need an inference model
        model_config.set_inference(inference=True)
        return ModelRegistry().init_model(
            model_config=model_config, string_replacement_map=string_replacement_map
        )

    def run(self) -> None:
        """
        Run the models specified in the ModelTimer config file, measure their inference time and log the results
        (inference time based on epoch time and bare process time as well as their means) to MLFlow.
        """
        model = ModelTimer.create_model(
            model_config=self.configuration.model_config,
            string_replacement_map=self.configuration.string_replacement_map,
        )

        image = cv2.imread(self.configuration.test_image_path)

        if self.mlflow_runner is not None:
            self.mlflow_runner.start_mlflow_run(
                experiment_name=MLFlowExperimentTypes.TIMING,
                run_name=model.unique_name,
            )
            logger.info("Log mlflow metrics for model '%s'", model.__class__)

        # TODO: Use the interface method in future implementations that state
        #       whether the model is running on gpu or cpu
        # mlflow.log_param(key="device_type", value="TODO")

        gpu_info: Optional[GpuInfo] = get_device_info(
            device_query=self.configuration.device_query,
        )

        if gpu_info is not None and mlflow.active_run() is not None:
            # log gpu info to mlflow
            mlflow.log_param(key="device_name", value=gpu_info.name)

        # perform warm-up
        for _ in range(0, self.configuration.number_of_warm_up_runs):
            start_epoch_time = time.time_ns()
            model.predict(data_item=image)
            logger.info("warm-up-time: %dms", time.time_ns() - start_epoch_time)

        # perform benchmarked inference
        for run_index in range(0, self.configuration.number_of_runs):
            start_epoch_time = time.time_ns()

            model.predict(data_item=image)

            self.inference_time_list.append(
                (time.time_ns() - start_epoch_time) / 1000 / 1000
            )

            if mlflow.active_run() is not None:
                mlflow.log_metric(
                    key="runtime_ms",
                    value=self.inference_time_list[-1],
                    step=run_index,
                )
            logger.info("runtime: %dms", self.inference_time_list[-1])

        if len(self.inference_time_list) > 0:
            self.inference_time_mean = sum(self.inference_time_list) / len(
                self.inference_time_list
            )
            logger.info("runtime-mean: %dms", self.inference_time_mean)
            if mlflow.active_run() is not None:
                mlflow.log_metric("runtime-mean_ms", value=self.inference_time_mean)

        if self.mlflow_runner is not None:
            self.mlflow_runner.end_run()


def main() -> None:
    """
    Entry point when using the mlcvzoo-modeltimer command line tool.
    (See [tool.poetry.scripts] section in pyproject.toml)
    """

    args = ConfigBuilder.setup_argparse(
        configure_argparse=ModelTimer.configure_argparse
    ).parse_args()

    Logger.init_logging_basic(
        log_dir=args.log_dir,
        log_file_postfix="ModelTimer",
        no_stdout=False,
        root_log_level=args.log_level,
    )

    configuration: ModelTimerConfig = ModelTimer.create_cli_configuration()

    configuration.model_config.update_class_type(
        args_dict=vars(args),
    )

    model_type: Optional[Type[Model[Any, Any, Any]]] = ModelRegistry().get_model_type(
        class_type=configuration.model_config.class_type
    )

    if model_type:
        configuration.model_config.update_constructor_parameters(
            args_dict=vars(args),
            model_type=model_type,
        )
    else:
        raise ValueError(
            f"Model type '{configuration.model_config.class_type}' is not registered. "
            f"Use one of '{list(ModelRegistry().get_registered_models().keys())}'",
        )

    model_timer = ModelTimer(configuration=configuration)
    model_timer.run()


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import logging
import os
from typing import Dict, List, Type
from unittest import main

from config_builder import BaseConfigClass, ConfigBuilder
from test_template import TestTemplate

from mlcvzoo_util.model_timer.configuration import ModelTimerConfig
from mlcvzoo_util.model_trainer.configuration import ModelTrainerConfig
from mlcvzoo_util.video_image_creator.configuration import VideoImageCreatorConfig

logger = logging.getLogger(__name__)


class TestTemplates(TestTemplate):
    def test_config_templates(self) -> None:
        template_path_dict: Dict[Type[BaseConfigClass], List[str]] = {
            VideoImageCreatorConfig: [
                os.path.join(
                    self.project_root,
                    "config",
                    "templates",
                    "tools",
                    "video-image-creator_template.yaml",
                )
            ],
            ModelTrainerConfig: [
                os.path.join(
                    self.project_root,
                    "config",
                    "templates",
                    "tools",
                    "model-trainer_config_template.yaml",
                )
            ],
            ModelTimerConfig: [
                os.path.join(
                    self.project_root,
                    "config",
                    "templates",
                    "tools",
                    "model-timer_config_template.yaml",
                )
            ],
        }

        for config_class_type, template_path_list in template_path_dict.items():
            for template_path in template_path_list:
                logger.info(
                    "=================================================================\n"
                    "CHECK correct build of configuration class %s "
                    "with template-config-path '%s'\n",
                    config_class_type,
                    template_path,
                )

                config_builder = ConfigBuilder(
                    class_type=config_class_type,
                    yaml_config_path=template_path,
                    string_replacement_map=self.string_replacement_map,
                    no_checks=True,
                )

                logger.info(
                    "================================================================="
                )
                assert config_builder.configuration is not None


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Tests for the LabelStudioTrackingTaskConverter
"""


import argparse
import logging
import os
from unittest import main, mock
from unittest.mock import MagicMock

from parametrize import parametrize
from test_template import TestTemplate

from mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter import (
    LabelStudioTrackingTaskConverter,
)
from mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter import (
    main as tracking_task_converter_main,
)

logger = logging.getLogger(__name__)


class TestLabelStudioTrackingTaskConverter(TestTemplate):
    """
    Tests for the LabelStudioTrackingTaskConverter
    """

    def setUp(self) -> None:
        """
        Setup file paths.
        """

        super().setUp()

        self.video_path = os.path.join(
            self.project_root,
            "test_data",
            "test_video.mp4",
        )
        self.tracking_annotations_path = os.path.join(
            self.project_root,
            "test_data",
            "test_tracking_task_converter",
            "tracking_annotations.json",
        )
        self.image_archive_path = os.path.join(
            self.project_root,
            "test_data",
            "test_tracking_task_converter",
            "image_archive.zip",
        )
        self.annotations_path = os.path.join(
            self.project_root,
            "test_data",
            "test_tracking_task_converter",
            "annotations.json",
        )

    def test_video_path_does_not_exist(self) -> None:
        """
        Test that an exception is raised, if the video file does not exist.
        """

        with self.assertRaisesRegex(ValueError, "Path to video '.*' does not exist."):
            LabelStudioTrackingTaskConverter(
                video_path=os.path.join(
                    self.project_root,
                    "test_data",
                    "test_video_invalid.mp4",
                ),
                tracking_annotations_path=self.tracking_annotations_path,
                image_archive_path=self.image_archive_path,
                annotations_path=self.annotations_path,
            )

    def test_tracking_annotations_path_does_not_exist(self) -> None:
        """
        Test that an exception is raised, if the file for the tracking annotations does not exist.
        """

        with self.assertRaisesRegex(
            ValueError, "Path to tracking annotations '.*' does not exist."
        ):
            LabelStudioTrackingTaskConverter(
                video_path=self.video_path,
                tracking_annotations_path=os.path.join(
                    self.project_root,
                    "test_data",
                    "test_tracking_task_converter",
                    "tracking_annotations_invalid.json",
                ),
                image_archive_path=self.image_archive_path,
                annotations_path=self.annotations_path,
            )

    def test_parse_video_specs_invalid_video(self) -> None:
        """
        Test that an exception is raised, if the video file is not valid.
        """

        with self.assertRaisesRegex(RuntimeError, "Could not open video '.*'"):
            tracking_converter = LabelStudioTrackingTaskConverter(
                video_path=self.tracking_annotations_path,
                tracking_annotations_path=self.tracking_annotations_path,
                image_archive_path=self.image_archive_path,
                annotations_path=self.annotations_path,
            )

            tracking_converter._parse_video_specs()

    def test_parse_video_specs(self) -> None:
        """
        Test that the video specs are parsed correctly.
        """

        tracking_converter = LabelStudioTrackingTaskConverter(
            video_path=self.video_path,
            tracking_annotations_path=self.tracking_annotations_path,
            image_archive_path=self.image_archive_path,
            annotations_path=self.annotations_path,
        )

        tracking_converter._parse_video_specs()

        self.assertEqual(tracking_converter.video_specs.width, 374)
        self.assertEqual(tracking_converter.video_specs.height, 500)
        self.assertEqual(tracking_converter.video_specs.frame_count, 10)
        self.assertEqual(tracking_converter.video_specs.frame_rate, 2.0)

    def test_build_image_filename(self) -> None:
        """
        Test that filenames for the image archive are built correctly.
        """

        tracking_converter = LabelStudioTrackingTaskConverter(
            video_path=self.video_path,
            tracking_annotations_path=self.tracking_annotations_path,
            image_archive_path=self.image_archive_path,
            annotations_path=self.annotations_path,
        )

        tracking_converter._parse_video_specs()

        self.assertEqual(
            tracking_converter._build_image_filename(0), "test_video_00.png"
        )
        self.assertEqual(
            tracking_converter._build_image_filename(9), "test_video_09.png"
        )

    @parametrize(
        "video_name",
        [
            ("test_video.mp4",),
            ("test_video_2.mp4",),
        ],
    )
    def test_find_task(self, video_name) -> None:
        """
        Test that the correct task is returned.
        """

        tracking_converter = LabelStudioTrackingTaskConverter(
            video_path=os.path.join(
                self.project_root,
                "test_data",
                video_name,
            ),
            tracking_annotations_path=self.tracking_annotations_path,
            image_archive_path=self.image_archive_path,
            annotations_path=self.annotations_path,
        )

        task = tracking_converter._find_task()
        self.assertEqual(task["file_upload"], video_name)

    @parametrize(
        "tracking_annotations_file,error_message_regex",
        [
            (
                "tracking_annotations_empty.json",
                "Invalid annotation file. Task for video 'test_video.mp4' is missing.",
            ),
            (
                "tracking_annotations_no_annotations.json",
                "Invalid task. Key annotations is missing or is empty.",
            ),
            (
                "tracking_annotations_no_result.json",
                "Invalid task. Key annotations.result is missing or is empty.",
            ),
            (
                "tracking_annotations_wrong_frame_count.json",
                "Invalid task.",
            ),
            (
                "tracking_annotations_no_labels.json",
                "Invalid task. Key annotations.result.value.labels is missing or is empty.",
            ),
            (
                "tracking_annotations_no_sequence.json",
                "Invalid task. Key annotations.result.value.sequence is missing or is empty.",
            ),
            (
                "tracking_annotations_no_file_upload.json",
                "Invalid task. Key file_upload is missing for a task.",
            ),
            (
                "tracking_annotations_task_missing.json",
                "Invalid annotation file. Task for video 'test_video.mp4' is missing.",
            ),
            (
                "tracking_annotations_wrong_annotation_format.json",
                "Annotation must contain keys 'frame', 'time', 'x', 'y', 'width', 'height' and "
                "'rotation'.",
            ),
        ],
    )
    def test_parse_tracking_annotations_file_invalid_annotations_file(
        self, tracking_annotations_file, error_message_regex
    ) -> None:
        """
        Test that exceptions are raised if the tracking annotations are invalid.
        """

        with self.assertRaisesRegex(ValueError, error_message_regex):
            tracking_converter = LabelStudioTrackingTaskConverter(
                video_path=self.video_path,
                tracking_annotations_path=os.path.join(
                    self.project_root,
                    "test_data",
                    "test_tracking_task_converter",
                    tracking_annotations_file,
                ),
                image_archive_path=self.image_archive_path,
                annotations_path=self.annotations_path,
            )

            tracking_converter._parse_video_specs()
            tracking_converter._parse_tracking_annotations_file()

    def test_parse_tracking_annotations_file(self) -> None:
        """
        Test that the tracking annotations are parsed correctly.
        """

        tracking_converter = LabelStudioTrackingTaskConverter(
            video_path=self.video_path,
            tracking_annotations_path=self.tracking_annotations_path,
            image_archive_path=self.image_archive_path,
            annotations_path=self.annotations_path,
        )

        tracking_converter._parse_video_specs()

        self.assertEqual(len(tracking_converter.annotations), 0)

        tracking_converter._parse_tracking_annotations_file()

        self.assertEqual(len(tracking_converter.annotations), 3)
        self.assertIn(0, tracking_converter.annotations)
        self.assertIn(1, tracking_converter.annotations)
        self.assertIn(2, tracking_converter.annotations)
        self.assertEqual(len(tracking_converter.annotations[0]), 2)
        self.assertEqual(len(tracking_converter.annotations[1]), 1)
        self.assertEqual(len(tracking_converter.annotations[2]), 1)

    @mock.patch(
        "mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter.json.dump"
    )
    def test_save_annotations(self, json_dump_mock) -> None:
        """
        Test that the parsed tracking annotations are saved correctly.
        """

        tracking_converter = LabelStudioTrackingTaskConverter(
            video_path=self.video_path,
            tracking_annotations_path=self.tracking_annotations_path,
            image_archive_path=self.image_archive_path,
            annotations_path=self.annotations_path,
        )

        # we want to mock open to prevent save_annotations to create a file
        # but open must return the correct data for the read in parse_annotations
        tracking_converter._parse_video_specs()
        tracking_converter._parse_tracking_annotations_file()
        with mock.patch(
            "mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter.open",
            mock.mock_open(),
        ):
            tracking_converter._save_annotations()

        data = json_dump_mock.call_args.kwargs["obj"]

        # assertCountEqual actually also checks equality of items
        self.assertCountEqual(
            data,
            [
                {
                    "annotations": [
                        {
                            "result": [
                                {
                                    "original_width": 374,
                                    "original_height": 500,
                                    "image_rotation": 0,
                                    "value": {
                                        "x": 42,
                                        "y": 42,
                                        "width": 42,
                                        "height": 42,
                                        "rotation": 0,
                                        "rectanglelabels": ["LabelA"],
                                    },
                                    "from_name": "label",
                                    "to_name": "image",
                                    "type": "rectanglelabels",
                                    "origin": "manual",
                                },
                                {
                                    "original_width": 374,
                                    "original_height": 500,
                                    "image_rotation": 0,
                                    "value": {
                                        "x": 84,
                                        "y": 84,
                                        "width": 84,
                                        "height": 84,
                                        "rotation": 0,
                                        "rectanglelabels": ["LabelB"],
                                    },
                                    "from_name": "label",
                                    "to_name": "image",
                                    "type": "rectanglelabels",
                                    "origin": "manual",
                                },
                            ]
                        }
                    ],
                    "data": {"image": "test_video_00.png"},
                },
                {
                    "annotations": [
                        {
                            "result": [
                                {
                                    "original_width": 374,
                                    "original_height": 500,
                                    "image_rotation": 0,
                                    "value": {
                                        "x": 42,
                                        "y": 42,
                                        "width": 42,
                                        "height": 42,
                                        "rotation": 0,
                                        "rectanglelabels": ["LabelA"],
                                    },
                                    "from_name": "label",
                                    "to_name": "image",
                                    "type": "rectanglelabels",
                                    "origin": "manual",
                                },
                            ]
                        }
                    ],
                    "data": {"image": "test_video_01.png"},
                },
                {
                    "annotations": [
                        {
                            "result": [
                                {
                                    "original_width": 374,
                                    "original_height": 500,
                                    "image_rotation": 0,
                                    "value": {
                                        "x": 84,
                                        "y": 84,
                                        "width": 84,
                                        "height": 84,
                                        "rotation": 0,
                                        "rectanglelabels": ["LabelB"],
                                    },
                                    "from_name": "label",
                                    "to_name": "image",
                                    "type": "rectanglelabels",
                                    "origin": "manual",
                                },
                            ]
                        }
                    ],
                    "data": {"image": "test_video_02.png"},
                },
            ],
        )

    @mock.patch(
        "mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter.zipfile.ZipFile",
        autospec=True,
    )
    @mock.patch(
        "mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter.cv2.imwrite"
    )
    def test_extract_images(self, cv2_imwrite_mock, zipfile_mock) -> None:
        """
        Test that the correct video frames are saved in the image archive.
        """

        zipfile_mock.return_value.__enter__.return_value.write = MagicMock()
        zipfile_write_mock = zipfile_mock.return_value.__enter__.return_value.write

        tracking_converter = LabelStudioTrackingTaskConverter(
            video_path=self.video_path,
            tracking_annotations_path=self.tracking_annotations_path,
            image_archive_path=self.image_archive_path,
            annotations_path=self.annotations_path,
        )

        tracking_converter._parse_video_specs()
        tracking_converter._parse_tracking_annotations_file()
        tracking_converter._extract_images()

        # assert cv2.imwrite
        self.assertEqual(len(cv2_imwrite_mock.call_args_list), 3)
        self.assertIn("test_video_00.png", cv2_imwrite_mock.call_args_list[0].args[0])
        self.assertIn("test_video_01.png", cv2_imwrite_mock.call_args_list[1].args[0])
        self.assertIn("test_video_02.png", cv2_imwrite_mock.call_args_list[2].args[0])

        # assert zipfile.ZipFile.write
        self.assertEqual(len(zipfile_write_mock.call_args_list), 3)
        self.assertIn("test_video_00.png", zipfile_write_mock.call_args_list[0].args[0])
        self.assertEqual(
            zipfile_write_mock.call_args_list[0].args[1], "test_video_00.png"
        )
        self.assertIn("test_video_01.png", zipfile_write_mock.call_args_list[1].args[0])
        self.assertEqual(
            zipfile_write_mock.call_args_list[1].args[1], "test_video_01.png"
        )
        self.assertIn("test_video_02.png", zipfile_write_mock.call_args_list[2].args[0])
        self.assertEqual(
            zipfile_write_mock.call_args_list[2].args[1], "test_video_02.png"
        )

    def test_extract_images_invalid_video(self) -> None:
        """
        Test that an exception is raised, if the video file is not valid.
        """

        with self.assertRaisesRegex(RuntimeError, "Could not open video '.*'"):
            tracking_converter = LabelStudioTrackingTaskConverter(
                video_path=self.tracking_annotations_path,
                tracking_annotations_path=self.tracking_annotations_path,
                image_archive_path=self.image_archive_path,
                annotations_path=self.annotations_path,
            )

            tracking_converter._extract_images()

    @mock.patch(
        "mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter"
        ".LabelStudioTrackingTaskConverter._parse_video_specs"
    )
    @mock.patch(
        "mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter"
        ".LabelStudioTrackingTaskConverter._parse_tracking_annotations_file"
    )
    @mock.patch(
        "mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter"
        ".LabelStudioTrackingTaskConverter._extract_images"
    )
    @mock.patch(
        "mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter."
        "LabelStudioTrackingTaskConverter._save_annotations"
    )
    def test_run(
        self,
        save_annotations_mock,
        extract_images_mock,
        parse_annotations_mock,
        get_video_specs_mock,
    ) -> None:
        """
        Test that the correct methods are called
        """

        tracking_converter = LabelStudioTrackingTaskConverter(
            video_path=self.video_path,
            tracking_annotations_path=self.tracking_annotations_path,
            image_archive_path=self.image_archive_path,
            annotations_path=self.annotations_path,
        )

        tracking_converter.run()

        self.assertEqual(get_video_specs_mock.called, True)
        self.assertEqual(parse_annotations_mock.called, True)
        self.assertEqual(extract_images_mock.called, True)
        self.assertEqual(save_annotations_mock.called, True)

    @mock.patch(
        "mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter"
        ".LabelStudioTrackingTaskConverter"
    )
    @mock.patch(
        "mlcvzoo_util.ls_tracking_task_converter.ls_tracking_task_converter.parse_args"
    )
    def test_main(
        self, parse_args_mock, label_studio_tracking_task_converter_mock
    ) -> None:
        """
        Test the main function instantiates the LabelStudioTrackingTaskConverter with the correct
        parameters.
        """

        parse_args_mock.return_value = argparse.Namespace(
            video_path=self.video_path,
            tracking_annotations_path=self.tracking_annotations_path,
            image_archive_path=self.image_archive_path,
            annotations_path=self.annotations_path,
        )

        tracking_task_converter_main()

        self.assertDictEqual(
            label_studio_tracking_task_converter_mock.call_args.kwargs,
            {
                "video_path": self.video_path,
                "tracking_annotations_path": self.tracking_annotations_path,
                "image_archive_path": self.image_archive_path,
                "annotations_path": self.annotations_path,
            },
        )


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

import copy
import logging
import os
import sys
from typing import Dict, Optional
from unittest import TestCase, main
from unittest.mock import MagicMock

from mlcvzoo_base.api.interfaces import NetBased, Trainable
from mlcvzoo_base.models.read_from_file.configuration import ReadFromFileConfig
from mlcvzoo_base.models.read_from_file.model import ReadFromFileObjectDetectionModel
from pytest import fixture, mark
from pytest_mock import MockerFixture
from test_template import TestTemplate

from mlcvzoo_util.model_trainer.model_trainer import ModelTrainer
from mlcvzoo_util.model_trainer.model_trainer import main as model_trainer_main

logger = logging.getLogger(__name__)


@fixture(scope="function")
def create_model_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch(
        "mlcvzoo_util.model_trainer.model_trainer.ModelTrainer.create_model",
        return_value=TestModel(
            from_yaml="",
        ),
    )


class TestModel(ReadFromFileObjectDetectionModel, NetBased, Trainable):
    def __init__(
        self,
        from_yaml: str,
        configuration: Optional[ReadFromFileConfig] = None,
        string_replacement_map: Optional[Dict[str, str]] = None,
        init_for_inference: bool = True,
    ):
        pass

    def get_training_output_dir(self) -> Optional[str]:
        return None

    def get_checkpoint_filename_suffix(self) -> str:
        return ""

    def restore(self, checkpoint_path: str) -> None:
        pass

    def train(self) -> None:
        pass

    def store(self, checkpoint_path: str) -> None:
        pass


class TestModelTrainer(TestTemplate):
    @mark.usefixtures(
        "create_model_mock",
    )
    def test_run_training_model_trainer_read_from_file(self) -> None:
        model_trainer = ModelTrainer(
            configuration=ModelTrainer.create_configuration(
                yaml_config_path=os.path.join(
                    self.project_root,
                    "test_data/test_model_trainer/"
                    "test_model-trainer_config_read-from-file_coco_test.yaml",
                ),
                string_replacement_map=self.string_replacement_map,
            )
        )

        model_trainer.run_training()

    @mark.usefixtures(
        "create_model_mock",
    )
    def test_model_trainer_main(self):
        argv_copy = copy.deepcopy(sys.argv)

        sys.argv = [sys.argv[0]]

        sys.argv.extend(
            [
                os.path.join(
                    self.project_root,
                    "test_data/test_model_trainer/"
                    "test_model-trainer_config_read-from-file_coco_test.yaml",
                ),
                "--replacement-config-path",
                self._gen_replacement_config(),
                "--log-dir",
                os.path.join(self.project_root, "test_output", "logs"),
                "--log-level",
                "DEBUG",
            ]
        )

        model_trainer_main()

        sys.argv = argv_copy


if __name__ == "__main__":
    main()



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

from typing import List, Optional

import related
from config_builder import BaseConfigClass
from mlcvzoo_base.configuration.device_query import ModelTimerDeviceQueryConfig
from mlcvzoo_base.configuration.mlfow_config import MLFlowConfig
from mlcvzoo_base.configuration.model_config import ModelConfig


@related.mutable(strict=True)
class ModelTimerConfig(BaseConfigClass):
    test_image_path: str = related.StringField(required=True)

    device_query: ModelTimerDeviceQueryConfig = related.ChildField(
        cls=ModelTimerDeviceQueryConfig,
    )

    model_config: ModelConfig = related.ChildField(cls=ModelConfig)

    mlflow_config: Optional[MLFlowConfig] = related.ChildField(
        cls=MLFlowConfig, required=False, default=None
    )

    number_of_runs: int = related.IntegerField(required=False, default=100)

    # Number of runs that are not accounted for determining the average runtime
    number_of_warm_up_runs: int = related.IntegerField(required=False, default=3)



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for configuring the parsing of information from yaml in python
accessible attributes for the PreAnnotationTool class
"""

import related
from config_builder import BaseConfigClass
from mlcvzoo_base.configuration.model_config import ModelConfig
from mlcvzoo_base.configuration.structs import OpenCVImageFormats

from mlcvzoo_util.cvat_annotation_handler.configuration import (
    CVATAnnotationHandlerConfig,
)


@related.mutable(strict=False)
class PreAnnotateCVATConfig(BaseConfigClass):
    """
    Main configuration class for the PreAnnotationTool
    """

    model_config: ModelConfig = related.ChildField(cls=ModelConfig)

    cvat_annotation_handler_config: CVATAnnotationHandlerConfig = related.ChildField(
        cls=CVATAnnotationHandlerConfig
    )

    root_image_dir: str = related.StringField()
    input_image_type: str = related.StringField()
    output_xml_dir: str = related.StringField()

    generate_annotations: bool = related.BooleanField()
    overwrite_existing_annotations: bool = related.BooleanField()

    show_predictions: bool = related.BooleanField(required=False, default=False)

    def check_values(self) -> bool:
        return True



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for configuring the parsing of information from yaml in python
accessible attributes for extracting frames from videos
"""
from typing import Dict

import related
from config_builder import BaseConfigClass


@related.mutable(strict=True)
class VideoImageCreatorConfig(BaseConfigClass):
    """Class for parsing information for video image creator tool"""

    default_step_width_map: Dict[str, int] = {
        "1": 1,
        "2": 2,
        "3": 3,
        "4": 4,
        "5": 5,
        "6": -1,
        "7": -2,
        "8": -3,
        "9": -4,
        "0": -5,
    }

    step_width_map: Dict[str, int] = related.ChildField(
        cls=dict, default=default_step_width_map
    )

    video_file_extension: str = related.StringField(default=".mp4")

    video_input_dir: str = related.StringField(default="")

    video_input_path: str = related.StringField(default="")

    winname: str = related.StringField(default="VideoImageCreator")

    window_size: int = related.IntegerField(default=500)
    resize_window: bool = related.BooleanField(default=True)



# Copyright Open Logistics Foundation
#
# Licensed under the Open Logistics Foundation License 1.3.
# For details on the licensing terms, see the LICENSE file.
# SPDX-License-Identifier: OLFL-1.3

"""
Module for handling the generation and uploading of annotations to CVAT, which realizes that
the feature of pre-annotating annotation tasks
"""

import argparse
import logging
import os
import zipfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Type, cast

import cv2
from config_builder import ConfigBuilder
from mlcvzoo_base.api.data.annotation import BaseAnnotation
from mlcvzoo_base.api.model import Model, ObjectDetectionModel
from mlcvzoo_base.configuration.model_config import ModelConfig
from mlcvzoo_base.data_preparation.utils import annotation_to_xml
from mlcvzoo_base.models.model_registry import ModelRegistry
from mlcvzoo_base.utils import draw_on_image, generate_detector_colors
from mlcvzoo_base.utils.file_utils import get_basename, get_file_list
from tqdm import tqdm

from mlcvzoo_util.cvat_annotation_handler.cvat_annotation_handler import (
    CVATAnnotationHandler,
)
from mlcvzoo_util.image_io_utils import VideoLiveOutput
from mlcvzoo_util.logger import Logger
from mlcvzoo_util.mlcvzoo_cli_tool import MLCVZooCLITool, configure_model_argparse
from mlcvzoo_util.pre_annotation_tool.configuration import PreAnnotateCVATConfig

logger = logging.getLogger(__name__)


class PreAnnotationTool(MLCVZooCLITool[PreAnnotateCVATConfig]):
    """
    Tool that generates predictions and uploads these to CVAT as initialization for new tasks
    """

    def __init__(self, configuration: PreAnnotateCVATConfig):
        """
        Instantiates a PreAnnotationTool object

        Args:
            configuration: The configuration for the PreAnnotationTool
        """

        self.configuration: PreAnnotateCVATConfig = configuration

        self.video_live_output: Optional[VideoLiveOutput] = None

    @staticmethod
    def create_configuration(
        yaml_config_path: str,
        string_replacement_map: Optional[Dict[str, str]] = None,
        no_checks: bool = False,
    ) -> PreAnnotateCVATConfig:
        """
        Create a PreAnnotateCVATConfig

        Args:
            yaml_config_path:  A yaml filepath where to build the configuration
                               object from
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
            no_checks: Whether the configuration object should be checked for mutual exclusiveness
                       and the "check_values" method for each attribute of the supertype
                       "BaseConfigClass" should be called
        Returns:
            The created configuration
        """

        return cast(
            PreAnnotateCVATConfig,
            ConfigBuilder(
                class_type=PreAnnotateCVATConfig,
                yaml_config_path=yaml_config_path,
                string_replacement_map=string_replacement_map,
                no_checks=no_checks,
            ).configuration,
        )

    @staticmethod
    def create_cli_configuration(
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> PreAnnotateCVATConfig:
        """
        Create a PreAnnotateCVATConfig

        Args:
            string_replacement_map: A dictionary that defines placeholders which can be used
                                    while parsing the file. They can be understood as variables
                                    that can be used to define configs that are valid across
                                    multiple devices.
        Returns:
            The created configuration
        """

        return MLCVZooCLITool._create_cli_configuration(
            configuration_class=PreAnnotateCVATConfig,
            string_replacement_map=string_replacement_map,
            configure_argparse=PreAnnotationTool.configure_argparse,
        )

    @staticmethod
    def configure_argparse(
        parser: argparse.ArgumentParser,
    ) -> None:
        parser.description = (
            "Generate predictions and upload these to "
            "CVAT as initialization for new tasks"
        )
        MLCVZooCLITool.configure_argparse(parser)
        configure_model_argparse(parser=parser)

    @staticmethod
    def create_model(
        model_config: ModelConfig,
        string_replacement_map: Optional[Dict[str, str]] = None,
    ) -> ObjectDetectionModel[Any, Any]:
        """
        Creates a model based on the given config

        Args:
            model_config: The model config that should be used
                          for creating a model
            string_replacement_map: (Optional) A dictionary that defines placeholders which can
                                    be used while parsing a configuration file. They can be
                                    understood as variables that can be used to define configs
                                    that are valid across multiple devices.
        Returns:
            The created model
        """
        # In order to be able to load the checkpoints that
        # should be evaluated, we need an inference model
        model_config.set_inference(inference=True)
        object_detection_model: Model[Any, Any, Any] = ModelRegistry().init_model(
            model_config=model_config, string_replacement_map=string_replacement_map
        )

        if not isinstance(object_detection_model, ObjectDetectionModel):
            raise ValueError(
                "This evaluation can only be used with models that "
                "inherit from 'mlcvzoo.api.model.ObjectDetectionModel'"
            )

        return object_detection_model

    def __gather_image_paths(self) -> List[str]:
        image_paths: List[str] = []

        disk_image_file_map: Dict[str, str] = {}

        input_image_types = self.configuration.input_image_type.split("|")

        input_image_paths: List[str] = []
        for input_image_type in input_image_types:
            input_image_paths.extend(
                get_file_list(
                    input_dir=self.configuration.root_image_dir,
                    search_subfolders=True,
                    file_extension=input_image_type,
                )
            )

        for image_path in input_image_paths:
            disk_image_file_map[get_basename(image_path)] = image_path

        for (
            dump_task_config
        ) in self.configuration.cvat_annotation_handler_config.dump_task_configs:
            source_zip_file = zipfile.ZipFile(dump_task_config.target_zip_path)

            logger.debug("Get image paths from '%s'" % dump_task_config.target_zip_path)

            task_image_paths = (
                source_zip_file.read("ImageSets/Main/default.txt")
                .decode("utf-8")
                .split("\n")
            )

            for task_image_path in task_image_paths:
                task_image_basename = get_basename(task_image_path)
                if task_image_basename in disk_image_file_map:
                    image_paths.append(disk_image_file_map[task_image_basename])

        logger.debug("Gathered '%s' image paths for prediction." % len(image_paths))

        return image_paths

    def generate_annotations(self) -> None:
        """
        Generate PASCAL-VOC annotations using the configured model and
        store these annotations at the configured location

        Returns:
            None
        """

        object_detection_model: ObjectDetectionModel[Any, Any] = (
            PreAnnotationTool.create_model(
                model_config=self.configuration.model_config,
                string_replacement_map=self.configuration.string_replacement_map,
            )
        )

        image_paths: List[str] = self.__gather_image_paths()

        process_bar = tqdm(image_paths, desc="Generate predictions")

        if self.configuration.show_predictions:
            self.video_live_output = VideoLiveOutput(mode=VideoLiveOutput.MODE_STEP)

        rgb_colors: List[Tuple[int, int, int]] = []
        if self.video_live_output is not None:
            rgb_colors = generate_detector_colors(
                num_classes=object_detection_model.num_classes
            )

        for image_path in process_bar:
            if (
                self.video_live_output is not None
                and self.video_live_output.is_terminated()
            ):
                break

            image = cv2.imread(image_path)

            predicted_annotation: BaseAnnotation = BaseAnnotation(
                image_path=image_path,
                annotation_path=os.path.join(
                    self.configuration.output_xml_dir, f"{get_basename(image_path)}.xml"
                ),
                image_shape=(image.shape[0], image.shape[1]),
                image_dir="",
                replacement_string="",
                annotation_dir=self.configuration.output_xml_dir,
                classifications=[],
                bounding_boxes=[],
                segmentations=[],
            )

            if not os.path.isfile(predicted_annotation.annotation_path) or (
                os.path.isfile(predicted_annotation.annotation_path)
                and self.configuration.overwrite_existing_annotations
            ):
                # TODO: When ReadFromFile model is supporting images, then switch
                #       to predict on image and not image-path
                _, predicted_bounding_boxes = object_detection_model.predict(
                    data_item=str(Path(image_path).resolve())
                )

                logger.info("Predicted bounding boxes: %s" % predicted_bounding_boxes)

                if self.video_live_output is not None:
                    self.video_live_output.output_frame(
                        draw_on_image(
                            frame=image,
                            rgb_colors=rgb_colors,
                            bounding_boxes=predicted_bounding_boxes,
                            thickness=5,
                        )
                    )

                predicted_annotation.bounding_boxes = predicted_bounding_boxes

                annotation_to_xml(
                    annotation=predicted_annotation,
                )

    def run(self) -> None:
        """
        Generate and upload the annotations to the CVAT instance

        1. Download all task information
        2. Parse the image-paths utilizing the information of the downloaded
           tasks
        3. Run the model based on the gathered image paths and generate the
           respective annotation files
        4. Uploaded the annotations to CVAT

        Returns:
            None
        """

        cvat_annotation_handler_instance = CVATAnnotationHandler(
            configuration=self.configuration.cvat_annotation_handler_config
        )

        cvat_annotation_handler_instance.download_all_tasks()

        if self.configuration.generate_annotations:
            self.generate_annotations()

        cvat_annotation_handler_instance.upload_all_tasks()


def main() -> None:
    """
    Entry point when using the mlcvzoo-preannotator command line tool.
    (See [tool.poetry.scripts] section in pyproject.toml)
    """

    args = ConfigBuilder.setup_argparse(
        configure_argparse=PreAnnotationTool.configure_argparse
    ).parse_args()

    Logger.init_logging_basic(
        log_dir=args.log_dir,
        log_file_postfix="PreAnnotationTool",
        no_stdout=False,
        root_log_level=args.log_level,
    )

    configuration: PreAnnotateCVATConfig = PreAnnotationTool.create_cli_configuration()

    configuration.model_config.update_class_type(
        args_dict=vars(args),
    )

    model_type: Optional[Type[Model[Any, Any, Any]]] = ModelRegistry().get_model_type(
        class_type=configuration.model_config.class_type
    )

    if model_type:
        configuration.model_config.update_constructor_parameters(
            args_dict=vars(args),
            model_type=model_type,
        )
    else:
        raise ValueError(
            f"Model type '{configuration.model_config.class_type}' is not registered. "
            f"Use one of '{list(ModelRegistry().get_registered_models().keys())}'",
        )

    pre_annotation_tool = PreAnnotationTool(configuration=configuration)
    pre_annotation_tool.run()


if __name__ == "__main__":
    main()
